from django.conf.urls import url

from . import views

urlpatterns = [
    url(r'^$', views.index, name='index'),
    url(r'^/(?P<testapp_id>\d+)$', views.show, name='show'),
    url(r'^/(?P<testapp_id>\d+)/urlendsinslash/$', views.slashed, name='slashed'),
    url(r'^/url\+with\+plus$', views.plused, name='plused'),
    url(r'^/quoted/(?P<slug>[^/]+)$', views.quoted, name='quoted'),
]



import re
from django.conf import settings
from django.core import urlresolvers
from django.core.exceptions import MiddlewareNotUsed
from django.http import HttpResponsePermanentRedirect, HttpResponseRedirect
from django.utils.encoding import iri_to_uri
try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:  # pragma: no cover
    MiddlewareMixin = object  # Django<1.10


trailing_slash_regexp = re.compile(r'(\/(?=\?))|(\/$)')


class RemoveSlashMiddleware(MiddlewareMixin):
    """
    This middleware provides the inverse of the APPEND_SLASH option built into
    django.middleware.common.CommonMiddleware. It should be placed just before
    or just after CommonMiddleware.

    If REMOVE_SLASH is True, the initial URL ends with a slash, and it is not
    found in the URLconf, then a new URL is formed by removing the slash at the
    end. If this new URL is found in the URLconf, then Django redirects the
    request to this new URL. Otherwise, the initial URL is processed as usual.

    For example, foo.com/bar/ will be redirected to foo.com/bar if you don't
    have a valid URL pattern for foo.com/bar/ but do have a valid pattern for
    foo.com/bar.

    Using this middleware with REMOVE_SLASH set to False or without
    REMOVE_SLASH set means it will do nothing.

    Originally, based closely on Django's APPEND_SLASH CommonMiddleware
    implementation at
    https://github.com/django/django/blob/master/django/middleware/common.py.
    It has been reworked to use regular expressions instead of deconstructing/
    reconstructing the URL, which was problematically re-encoding some of the
    characters.
    """

    def __init__(self, *args, **kwargs):
        if not getattr(settings, 'REMOVE_SLASH', False):
            raise MiddlewareNotUsed()
        self.response_class = HttpResponsePermanentRedirect
        if getattr(settings, 'UNSLASHED_USE_302_REDIRECT', None):
            self.response_class = HttpResponseRedirect
        super(RemoveSlashMiddleware, self).__init__(*args, **kwargs)

    def should_redirect_without_slash(self, request):
        """
        Return True if appending a slash to
        the request path turns an invalid path into a valid one.
        """
        if trailing_slash_regexp.search(request.get_full_path()):
            urlconf = getattr(request, 'urlconf', None)

            return (
                not urlresolvers.is_valid_path(request.path_info, urlconf) and
                urlresolvers.is_valid_path(request.path_info[:-1], urlconf))
        return False

    def get_full_path_without_slash(self, request):
        """
        Return the full path of the request without a trailing slash appended.

        Raise a RuntimeError if settings.DEBUG is True and request.method is
        POST, PUT, or PATCH.
        """
        path = request.path
        if path.endswith('/'):
            path = path[:-1]
        new_path = '{}{}'.format(
            path,
            ('?' + iri_to_uri(request.META.get('QUERY_STRING', '')))
            if request.META.get('QUERY_STRING', '')
            else ''
        )
        if settings.DEBUG and request.method in ('POST', 'PUT', 'PATCH'):
            raise RuntimeError(
                "You called this URL via %(method)s, but the URL ends "
                "in a slash and you have REMOVE_SLASH set. Django-unslashed "
                "can't redirect to the slash URL while maintaining %(method)s "
                "data. Change your form to point to %(url)s (note there is no"
                "trailing slash), or set REMOVE_SLASH=False in your Django "
                " settings." % {
                    'method': request.method,
                    'url': request.get_host() + new_path,
                }
            )
        return new_path

    def process_response(self, request, response):
        """
        Redirects to the current URL without the trailing slash if
        settings.REMOVE_SLASH is True and our current response's
        status_code would be a 404
        """
        # If the given URL is "Not Found", then check if we should redirect to
        # a path without a slash appended.
        if response.status_code == 404:
            path = request.path
            whitelists = getattr(settings,
                                 'UNSLASHED_WHITELIST_STARTSWITH',
                                 ['/admin'])
            if any(path.startswith(x) for x in whitelists):
                return response
            if self.should_redirect_without_slash(request):
                return self.response_class(
                    self.get_full_path_without_slash(request))
        return response



from django.http import HttpResponse

def index(request):
    return HttpResponse("Testapp index listing.")

def show(request, testapp_id):
    return HttpResponse("Showing testapp %s" % testapp_id)

def slashed(request, testapp_id):
    return HttpResponse("Showing testapp %s at url that should be / terminated." % testapp_id)

def plused(request):
    return HttpResponse("Showing URLs with plus signs are not re-encoded")

def quoted(request):
    return HttpResponse("Showing URLs with url-encoding are not re-encoded")



from django.conf.urls import include, url

# from django.contrib import admin
# admin.autodiscover()

urlpatterns = [
    # Examples:
    # url(r'^blog/', include('blog.urls')),
    # url(r'^admin/', include(admin.site.urls)),
    url(r'^testapps', include('testapps.urls', namespace='testapps')),
]



import os
from setuptools import setup
from unslashed import __version__
import pypandoc

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the relevant file
long_description = pypandoc.convert('README.md', 'rst')

setup(
    name='tocka-django-unslashed',

    # Versions should comply with PEP440. For single-sourced versioning, see
    # https://packaging.python.org/en/latest/development.html#single-sourcing-the-version
    version=__version__,

    description='Django Middleware that can automatically remove trailing URL slashes and 301 redirect to the non-slash-terminated URL.',
    long_description=long_description,

    # The project URL
    url = 'https://github.com/frnhr/django-unslashed',

    # Author details
    author='Fran Hrzenjak',
    author_email='fran.hrzenjak@gmail.com',

    # Choose your license
    license='MIT License',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        # Extras
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Framework :: Django',
        'Environment :: Web Environment',
        'Topic :: Internet :: WWW/HTTP'],

    # What does your project relate to?
    keywords='django slash remove trailing unslash remove_slash path',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=['unslashed'],

    # Run-time package dependencies. These will be installed by pip when your
    # project is installed.
    install_requires=['Django>=1.8'],

    # MANIFEST.in included entries should be included as package data and
    # installed into site-packages 
    include_package_data=True,

    # Default to False unless you specifically intend the package to be
    # installed as an .egg
    zip_safe=False,
)


from django.http.response import (
    HttpResponse,
    HttpResponsePermanentRedirect,
    HttpResponseRedirect,
    HttpResponseRedirectBase,
)
from django.test import TestCase, Client
from django.utils.http import urlquote


class RemoveSlashMiddlewareTest(TestCase):
    def setUp(self):
        self.client = Client()

    def test_permanent_redirect_to_non_slashed(self):
        response = self.client.get('/testapps/', follow=False)
        self.assertIsInstance(response, HttpResponsePermanentRedirect)
        self.assertFalse(response['Location'].endswith('/'))

        response = self.client.get('/testapps/1/', follow=False)
        self.assertIsInstance(response, HttpResponsePermanentRedirect)
        self.assertFalse(response['Location'].endswith('/'))

    def test_temp_redirect_to_non_slashed(self):
        with self.settings(UNSLASHED_USE_302_REDIRECT=True):
            response = self.client.get('/testapps/', follow=False)
            self.assertIsInstance(response, HttpResponseRedirect)
            self.assertFalse(response['Location'].endswith('/'))

            response = self.client.get('/testapps/1/', follow=False)
            self.assertIsInstance(response, HttpResponseRedirect)
            self.assertFalse(response['Location'].endswith('/'))

    def test_whitelisted_slashed(self):
        with self.settings(UNSLASHED_WHITELIST_STARTSWITH=['/testapps/']):
            response = self.client.get('/testapps/', follow=False)
            self.assertNotIsInstance(response, HttpResponseRedirectBase)
            self.assertEqual(response.status_code, 404)

            response = self.client.get('/testapps/1/', follow=False)
            self.assertNotIsInstance(response, HttpResponseRedirectBase)
            self.assertEqual(response.status_code, 404)

    def test_permanent_redirect_to_unslashed_when_url_has_plus_signs(self):
        response = self.client.get('/testapps/url+with+plus/', follow=False)
        self.assertIsInstance(response, HttpResponsePermanentRedirect)
        self.assertFalse(response['Location'].endswith('/'))
        self.assertRegexpMatches(response['Location'], r'/url\+with\+plus$')

        response = self.client.get('/testapps/url+with+plus/?param1=1&param2=%2Bfoo', follow=False)
        self.assertIsInstance(response, HttpResponsePermanentRedirect)
        self.assertRegexpMatches(response['Location'], r'/url\+with\+plus\?param1=1&param2=%2Bfoo$')

    def test_permanent_redirect_to_unslashed_when_url_has_urlencoded_chars(self):
        # urlquote is used because django.test.Client decodes the URL
        # which goes against the test
        slashed_url = urlquote('/testapps/quoted/foo%2Bbar%23baz%20/')
        response = self.client.get(slashed_url, follow=False)
        self.assertIsInstance(response, HttpResponsePermanentRedirect)
        self.assertFalse(response['Location'].endswith('/'))
        self.assertRegexpMatches(response['Location'], r'/quoted/foo%2Bbar%23baz%20$')

        response = self.client.get(slashed_url + '?param1=1&param2=%2Btest', follow=False)
        self.assertIsInstance(response, HttpResponsePermanentRedirect)
        self.assertRegexpMatches(response['Location'], r'/quoted/foo%2Bbar%23baz%20\?param1=1&param2=%2Btest$')

    def test_no_redirect_when_slash_url_is_valid(self):
        response = self.client.get('/testapps/1/urlendsinslash/', follow=False)
        self.assertIsInstance(response, HttpResponse)
        self.assertEqual(response.status_code, 200)

    def test_no_redirect_when_slashed_and_unslashed_invalid(self):
        response = self.client.get('/testapps/invalid/', follow=False)
        self.assertNotIsInstance(response, HttpResponseRedirectBase)
        self.assertEqual(response.status_code, 404)

        response = self.client.get('/testapps/invalid', follow=False)
        self.assertNotIsInstance(response, HttpResponseRedirectBase)
        self.assertEqual(response.status_code, 404)

    def test_warns_about_redirect_and_post(self):
        with self.settings(DEBUG=True):
            with self.assertRaises(RuntimeError):
                self.client.post('/testapps/', follow=False)

    def test_disabled_middleware(self):
        with self.settings(REMOVE_SLASH=False):
            response = self.client.get('/testapps/', follow=False)
            self.assertNotIsInstance(response, HttpResponseRedirectBase)
            self.assertEqual(response.status_code, 404)

